@@ -48,6 +48,7 @@ pub struct NetworkInfo {
 }
 
 /// Builder to construct client instance with sensible default values
+#[derive(Clone)]
 pub struct ClientBuilder {
     node_manager_builder: crate::node_manager::NodeManagerBuilder,
     #[cfg(not(feature = "wasm"))]
@@ -212,8 +213,9 @@ impl ClientBuilder {
     }
 
     /// Selects the type of network to get default nodes for it, only "testnet" is supported at the moment.
-    /// Nodes that don't belong to this network are ignored. Default nodes are only used when no other nodes are
-    /// provided.
+    /// Nodes that don't belong to this network are ignored. The &str must match a part or all of the networkId returned
+    /// in the nodeinfo from a node. For example, if the networkId is `"private-tangle"`, `"tangle"` can be used.
+    /// Default nodes are only used when no other nodes are provided.
     pub fn with_network(mut self, network: &str) -> Self {
         self.network_info.network.replace(network.into());
         self
@@ -260,12 +262,10 @@ impl ClientBuilder {
                 return Err(Error::MissingParameter("Node"));
             }
         }
-
         let network_info = Arc::new(RwLock::new(self.network_info));
         let nodes = self.node_manager_builder.nodes.clone();
         #[cfg(not(feature = "wasm"))]
         let node_sync_interval = self.node_sync_interval;
-
         #[cfg(feature = "wasm")]
         let (sync, network_info) = (Arc::new(RwLock::new(nodes)), network_info);
         #[cfg(not(feature = "wasm"))]
