@@ -12,7 +12,10 @@ use crate::{
 use bee_common::packable::Packable;
 use bee_message::{
     constants::INPUT_OUTPUT_COUNT_MAX,
-    prelude::{Address, Ed25519Address, Message, MessageBuilder, MessageId, Parents, TransactionId, UtxoInput},
+    prelude::{
+        Address, Ed25519Address, Message, MessageBuilder, MessageId, Parents, TransactionId, UtxoInput,
+        ED25519_ADDRESS_LENGTH,
+    },
 };
 use bee_pow::providers::{
     miner::{MinerBuilder, MinerCancel},
@@ -36,7 +39,6 @@ use crypto::{
 
 use zeroize::Zeroize;
 
-#[cfg(not(feature = "wasm"))]
 use crate::builder::TIPS_INTERVAL;
 #[cfg(feature = "mqtt")]
 use rumqttc::AsyncClient as MqttClient;
@@ -146,7 +148,7 @@ fn default_broker_port() -> u16 {
 
 #[cfg(feature = "mqtt")]
 fn default_max_reconnection_attempts() -> usize {
-    3
+    0
 }
 
 #[cfg(feature = "mqtt")]
@@ -193,7 +195,7 @@ impl BrokerOptions {
         self
     }
 
-    /// Sets the maximum number of reconnection attempts.
+    /// Sets the maximum number of reconnection attempts. 0 is unlimited.
     pub fn max_reconnection_attempts(mut self, max_reconnection_attempts: usize) -> Self {
         self.max_reconnection_attempts = max_reconnection_attempts;
         self
@@ -260,7 +262,7 @@ impl NonceProvider for ClientMiner {
 }
 
 /// Each of the node APIs the client uses.
-#[derive(Eq, PartialEq, Hash)]
+#[derive(Clone, Eq, PartialEq, Hash)]
 pub enum Api {
     /// `get_health` API
     GetHealth,
@@ -306,6 +308,7 @@ impl FromStr for Api {
 }
 
 /// An instance of the client using HORNET or Bee URI
+#[cfg_attr(feature = "wasm", derive(Clone))]
 pub struct Client {
     #[allow(dead_code)]
     #[cfg(not(feature = "wasm"))]
@@ -498,7 +501,7 @@ impl Client {
             let info = self.get_info().await?.nodeinfo;
             let network_id = hash_network(&info.network_id).ok();
             {
-                let mut client_network_info = self.network_info.write().map_err(|e| e).unwrap();
+                let mut client_network_info = self.network_info.write().map_err(|_| crate::Error::PoisonError)?;
                 client_network_info.network_id = network_id;
                 client_network_info.min_pow_score = info.min_pow_score;
                 client_network_info.bech32_hrp = info.bech32_hrp;
@@ -522,7 +525,6 @@ impl Client {
     }
 
     /// returns the tips interval
-    #[cfg(not(feature = "wasm"))]
     pub async fn get_tips_interval(&self) -> u64 {
         self.network_info
             .read()
@@ -564,7 +566,7 @@ impl Client {
         crypto::keys::bip39::wordlist::verify(mnemonic, &crypto::keys::bip39::wordlist::ENGLISH)
             .map_err(|e| crate::Error::InvalidMnemonic(format!("{:?}", e)))?;
         let mut mnemonic_seed = [0u8; 64];
-        mnemonic_to_seed(mnemonic, &"", &mut mnemonic_seed);
+        mnemonic_to_seed(mnemonic, "", &mut mnemonic_seed);
         Ok(hex::encode(mnemonic_seed))
     }
 
@@ -762,11 +764,76 @@ impl Client {
             #[serde(rename = "messageId")]
             message_id: String,
         }
+
+        #[cfg(not(feature = "pow-fallback"))]
         let resp: ResponseWrapper = self
             .node_manager
             .post_request_bytes(path, timeout, &message.pack_new(), local_pow)
             .await?;
 
+        #[cfg(feature = "pow-fallback")]
+        // fallback to local PoW if remote PoW fails
+        let resp: ResponseWrapper = match self
+            .node_manager
+            .post_request_bytes(path, timeout, &message.pack_new(), local_pow)
+            .await
+        {
+            Ok(res) => res,
+            Err(e) => {
+                if let Error::NodeError(e) = e {
+                    // hornet and bee return different error messages
+                    if e == *"No available nodes with remote PoW"
+                        || e.contains("proof of work is not enabled")
+                        || e.contains("`PoW` not enabled")
+                    {
+                        let mut client_network_info =
+                            self.network_info.write().map_err(|_| crate::Error::PoisonError)?;
+                        // switch to local PoW
+                        client_network_info.local_pow = true;
+                        drop(client_network_info);
+                        #[cfg(not(feature = "wasm"))]
+                        let msg_res = crate::api::finish_pow(self, message.payload().clone()).await;
+                        #[cfg(feature = "wasm")]
+                        let msg_res = {
+                            let min_pow_score = self.get_min_pow_score().await?;
+                            let network_id = self.get_network_id().await?;
+                            crate::api::finish_single_thread_pow(
+                                self,
+                                network_id,
+                                None,
+                                message.payload().clone(),
+                                min_pow_score,
+                            )
+                            .await
+                        };
+                        let message_with_local_pow = match msg_res {
+                            Ok(msg) => {
+                                // reset local PoW state
+                                let mut client_network_info =
+                                    self.network_info.write().map_err(|_| crate::Error::PoisonError)?;
+                                client_network_info.local_pow = false;
+                                msg
+                            }
+                            Err(e) => {
+                                // reset local PoW state
+                                let mut client_network_info =
+                                    self.network_info.write().map_err(|_| crate::Error::PoisonError)?;
+                                client_network_info.local_pow = false;
+                                return Err(e);
+                            }
+                        };
+                        self.node_manager
+                            .post_request_bytes(path, timeout, &message_with_local_pow.pack_new(), true)
+                            .await?
+                    } else {
+                        return Err(Error::NodeError(e));
+                    }
+                } else {
+                    return Err(e);
+                }
+            }
+        };
+
         let mut message_id_bytes = [0u8; 32];
         hex::decode_to_slice(resp.data.message_id, &mut message_id_bytes)?;
         Ok(MessageId::from(message_id_bytes))
@@ -781,7 +848,7 @@ impl Client {
         } else {
             self.get_timeout(Api::PostMessageWithRemotePow)
         };
-        let message = MessageDto::from(message);
+        let message_dto = MessageDto::from(message);
         #[derive(Debug, Serialize, Deserialize)]
         struct ResponseWrapper {
             data: MessageIdWrapper,
@@ -792,10 +859,69 @@ impl Client {
             message_id: String,
         }
 
-        let resp: ResponseWrapper = self
+        // fallback to local PoW if remote PoW fails
+        let resp: ResponseWrapper = match self
             .node_manager
-            .post_request_json(path, timeout, serde_json::to_value(message)?, local_pow)
-            .await?;
+            .post_request_json(path, timeout, serde_json::to_value(message_dto)?, local_pow)
+            .await
+        {
+            Ok(res) => res,
+            Err(e) => {
+                if let Error::NodeError(e) = e {
+                    // hornet and bee return different error messages
+                    if e == *"No available nodes with remote PoW"
+                        || e.contains("proof of work is not enabled")
+                        || e.contains("`PoW` not enabled")
+                    {
+                        let mut client_network_info =
+                            self.network_info.write().map_err(|_| crate::Error::PoisonError)?;
+                        // switch to local PoW
+                        client_network_info.local_pow = true;
+                        drop(client_network_info);
+                        #[cfg(not(feature = "wasm"))]
+                        let msg_res = crate::api::finish_pow(self, message.payload().clone()).await;
+                        #[cfg(feature = "wasm")]
+                        let msg_res = {
+                            let min_pow_score = self.get_min_pow_score().await?;
+                            let network_id = self.get_network_id().await?;
+                            crate::api::finish_single_thread_pow(
+                                self,
+                                network_id,
+                                None,
+                                message.payload().clone(),
+                                min_pow_score,
+                            )
+                            .await
+                        };
+                        let message_with_local_pow = match msg_res {
+                            Ok(msg) => {
+                                // reset local PoW state
+                                let mut client_network_info =
+                                    self.network_info.write().map_err(|_| crate::Error::PoisonError)?;
+                                client_network_info.local_pow = false;
+                                msg
+                            }
+                            Err(e) => {
+                                // reset local PoW state
+                                let mut client_network_info =
+                                    self.network_info.write().map_err(|_| crate::Error::PoisonError)?;
+                                client_network_info.local_pow = false;
+                                return Err(e);
+                            }
+                        };
+                        let message_dto = MessageDto::from(&message_with_local_pow);
+
+                        self.node_manager
+                            .post_request_json(path, timeout, serde_json::to_value(message_dto)?, true)
+                            .await?
+                    } else {
+                        return Err(Error::NodeError(e));
+                    }
+                } else {
+                    return Err(e);
+                }
+            }
+        };
 
         let mut message_id_bytes = [0u8; 32];
         hex::decode_to_slice(resp.data.message_id, &mut message_id_bytes)?;
@@ -843,7 +969,7 @@ impl Client {
         // Use `get_address()` API to get the address outputs first,
         // then collect the `UtxoInput` in the HashSet.
         for address in addresses {
-            let address_outputs = self.get_address().outputs(&address, Default::default()).await?;
+            let address_outputs = self.get_address().outputs(address, Default::default()).await?;
             for output in address_outputs.iter() {
                 output_to_query.insert(output.to_owned());
             }
@@ -853,7 +979,7 @@ impl Client {
                 let address_dust_allowance_outputs = self
                     .get_address()
                     .outputs(
-                        &address,
+                        address,
                         OutputsOptions {
                             include_spent: false,
                             output_type: Some(OutputType::SignatureLockedDustAllowance),
@@ -996,7 +1122,12 @@ impl Client {
             #[cfg(feature = "wasm")]
             {
                 let network_id = self.get_network_id().await?;
-                let mut message_builder = MessageBuilder::<ClientMiner>::new().with_network_id(network_id);
+                let mut tips = self.get_tips().await?;
+                tips.sort_unstable_by_key(|a| a.pack_new());
+                tips.dedup();
+                let mut message_builder = MessageBuilder::<ClientMiner>::new()
+                    .with_network_id(network_id)
+                    .with_parents(Parents::new(tips)?);
                 if let Some(p) = message.payload().to_owned() {
                     message_builder = message_builder.with_payload(p)
                 }
@@ -1072,7 +1203,7 @@ impl Client {
 
     /// Return a list of addresses from the seed regardless of their validity.
     pub fn get_addresses<'a>(&'a self, seed: &'a Seed) -> GetAddressesBuilder<'a> {
-        GetAddressesBuilder::new(seed).with_client(&self)
+        GetAddressesBuilder::new(seed).with_client(self)
     }
 
     /// Find all messages by provided message IDs and/or indexation_keys.
@@ -1115,12 +1246,12 @@ impl Client {
         GetBalanceBuilder::new(self, seed)
     }
 
-    /// Return the balance in iota for the given addresses; No seed or security level needed to do this
-    /// since we are only checking and already know the addresses.
+    /// Return the balance in iota for the given addresses; No seed needed to do this since we are only checking and
+    /// already know the addresses.
     pub async fn get_address_balances(&self, addresses: &[String]) -> Result<Vec<BalanceAddressResponse>> {
         let mut address_balance_pairs = Vec::new();
         for address in addresses {
-            let balance_response = self.get_address().balance(&address).await?;
+            let balance_response = self.get_address().balance(address).await?;
             address_balance_pairs.push(balance_response);
         }
         Ok(address_balance_pairs)
@@ -1133,7 +1264,7 @@ impl Client {
         Ok(ed.to_string())
     }
 
-    /// Transforms hex to bech32
+    /// Transforms a hex encoded address to a bech32 encoded address
     pub async fn hex_to_bech32(&self, hex: &str, bech32_hrp: Option<&str>) -> crate::Result<String> {
         let address: Ed25519Address = hex.parse::<Ed25519Address>()?;
         match bech32_hrp {
@@ -1142,12 +1273,27 @@ impl Client {
         }
     }
 
+    /// Transforms a hex encoded public key to a bech32 encoded address
+    pub async fn hex_public_key_to_bech32_address(&self, hex: &str, bech32_hrp: Option<&str>) -> crate::Result<String> {
+        let mut public_key = [0u8; ED25519_ADDRESS_LENGTH];
+        hex::decode_to_slice(&hex, &mut public_key)?;
+
+        let address = Blake2b256::digest(&public_key)
+            .try_into()
+            .map_err(|_e| Error::Blake2b256Error("Hashing the public key failed."))?;
+        let address: Ed25519Address = Ed25519Address::new(address);
+        match bech32_hrp {
+            Some(hrp) => Ok(Address::Ed25519(address).to_bech32(hrp)),
+            None => Ok(Address::Ed25519(address).to_bech32(self.get_bech32_hrp().await?.as_str())),
+        }
+    }
+
     /// Returns a valid Address parsed from a String.
     pub fn parse_bech32_address(address: &str) -> crate::Result<Address> {
         Ok(Address::try_from_bech32(address)?)
     }
 
-    /// Checks if a String address is valid.
+    /// Checks if a String is a valid bech32 encoded address.
     pub fn is_address_valid(address: &str) -> bool {
         Address::try_from_bech32(address).is_ok()
     }
@@ -1167,7 +1313,7 @@ impl Client {
     }
 
     /// Retries (promotes or reattaches) a message for provided message id until it's included (referenced by a
-    /// milestone). Default interval is 5 seconds and max attempts is 10. Returns reattached messages
+    /// milestone). Default interval is 5 seconds and max attempts is 40. Returns reattached messages
     pub async fn retry_until_included(
         &self,
         message_id: &MessageId,
@@ -1178,15 +1324,18 @@ impl Client {
         let mut message_ids = vec![*message_id];
         // Reattached Messages that get returned
         let mut messages_with_id = Vec::new();
-        for _ in 0..max_attempts.unwrap_or(20) {
+        for _ in 0..max_attempts.unwrap_or(40) {
             #[cfg(feature = "wasm")]
-            std::thread::sleep(Duration::from_secs(interval.unwrap_or(5)));
+            {
+                use wasm_timer::Delay;
+                Delay::new(Duration::from_secs(interval.unwrap_or(5))).await?;
+            }
             #[cfg(not(feature = "wasm"))]
             sleep(Duration::from_secs(interval.unwrap_or(5))).await;
             // Check inclusion state for each attachment
             let message_ids_len = message_ids.len();
             for (index, msg_id) in message_ids.clone().iter().enumerate() {
-                let message_metadata = self.get_message().metadata(&msg_id).await?;
+                let message_metadata = self.get_message().metadata(msg_id).await?;
                 if message_metadata.ledger_inclusion_state.is_some() {
                     return Ok(messages_with_id);
                 }
@@ -1194,10 +1343,10 @@ impl Client {
                 if index == message_ids_len - 1 {
                     if message_metadata.should_promote.unwrap_or(false) {
                         // Safe to unwrap since we iterate over it
-                        self.promote_unchecked(&message_ids.last().unwrap()).await?;
+                        self.promote_unchecked(message_ids.last().unwrap()).await?;
                     } else if message_metadata.should_reattach.unwrap_or(false) {
                         // Safe to unwrap since we iterate over it
-                        let reattached = self.reattach_unchecked(&message_ids.last().unwrap()).await?;
+                        let reattached = self.reattach_unchecked(message_ids.last().unwrap()).await?;
                         message_ids.push(reattached.0);
                         messages_with_id.push(reattached);
                     }
@@ -1215,7 +1364,7 @@ impl Client {
         account_index: usize,
         address_range: Range<usize>,
     ) -> crate::Result<String> {
-        crate::api::consolidate_funds(&self, seed, account_index, address_range).await
+        crate::api::consolidate_funds(self, seed, account_index, address_range).await
     }
 }
 
